'''
Created on Mar 8, 2016

@author: mukuze
'''
import wx
import wx.grid as gridlib
import os
import cPickle
import numpy as np

# imports to be able to implement threads. The recommending next operation takes too
# much time so it is better to implement it as a separate thread
import threading

# import modules
import Exact_bool_factorization as ebf
import Constraints_new as cnstr
import Visualization_functions as vs
import plot_GUI as visualization

# import the module containing the global variables that we want to share across modules
import important_data as dt

# some global variable
path_input_f=""
path_input_=""
inp=[]
trans_1s=[] # transactions with their items. We use this to calculate the number of 1s in each row
cols_1s={} # dictionary which gives the number of 1s of each column in the data
_rows=0
_cols=0
delete_list=[] # list of factors to delete. The user can select one or more factors to delete
curr_row="" # currently selected row
curr_rows=[] # currently selected rows
merge_pair=[] # pair to merge
count=0 

# global variable to count how many times the user has rejected the recommended operation
skip_count=0

####
# DIFFERENT THREAD
####   

# define notification event for thread completion
EVT_RESULT_ID=wx.NewId()
 
def EVT_RESULT(win, func):
     
    # define Result event
     
    win.Connect(-1, -1, EVT_RESULT_ID, func)
     
# simple event to carry the result data
class ResultEvent(wx.PyEvent):
    def __init__(self, data):
        # initialize Result event
        wx.PyEvent.__init__(self)
        self.SetEventType(EVT_RESULT_ID)
        self.data=data
         
# thread class that computes the recommendation of best operation computation
class RecommendThread(threading.Thread):
     
    def __init__(self, notify_window, input_mat, indices_1s, nb_cols, nb_rows, closed, trans, factors, scores):
        # initialize Thread class
        threading.Thread.__init__(self)
         
        self._notify_window=notify_window
         
        # initialize the class variables
        self.input_mat=input_mat
        self.indices_1s=indices_1s
        self.nb_cols=nb_cols
        self.nb_rows=nb_rows
        self.closed=closed
        self.trans=trans
        self.factors=factors
        self.scores=scores
         
    def run(self):
        # after the data is loaded we can now recommend the best operation to the user
        # for this purpose we call the cnstr.best_action_perform function in the Constraints_new module  
        # first, set some variables in the Constraints_new module
        cnstr.get_set_values(self.input_mat, self.indices_1s, self.nb_cols, self.nb_rows)
            
        # then call the function that returns the recommended operation
        recommend_op=cnstr.best_action_perform(self.closed, self.trans, self.factors, self.scores)
             
        wx.PostEvent(self._notify_window, ResultEvent(recommend_op))
                
####
# we define the different panels on the first tab
# tab1
# -------
# Top Panel (Here we show the factors and the info about them
# -------
# Low Panel (Here we allow the user to make some deletions/merges and see recommended operations
####

####
# Top Panel
####
class TopPanel(wx.Panel):
     
    def __init__(self, parent):
        wx.Panel.__init__(self, parent=parent)
         
#### 
# Low Panel
####
class LowPanel(wx.Panel):
     
    def __init__(self, parent):
        wx.Panel.__init__(self, parent=parent)
        
        #####
        # CONTROLS
        #####
        # Add controls
        # Delete
        self.Delete_txt=wx.StaticText(self, -1, "Delete Operation")
        self.Delete_error_txt=wx.StaticText(self, -1, "Error:")
        self.Delete_error=wx.StaticText(self, -1, "")
        self.Delete_btn=wx.Button(self, 101, 'Delete')
        
        # Merge
        self.Merge_txt=wx.StaticText(self, -1, "Merge Operation")
        self.Merge_OR_error_txt=wx.StaticText(self, -1, "Error:")
        self.Merge_OR_error=wx.StaticText(self, -1, "")
        self.Merge_AND_error_txt=wx.StaticText(self, -1, "Error:")
        self.Merge_AND_error=wx.StaticText(self, -1, "")
        self.Merge_Visualize_btn=wx.Button(self, 102, 'Visualize')
        self.Merge_OR_btn=wx.Button(self, 103, 'Merge OR')
        self.Merge_AND_btn=wx.Button(self, 104, 'Merge AND')
        
        # Statistics
        self.statistics_txt=wx.StaticText(self, -1, "Statistics")
        self.number_rows_txt=wx.StaticText(self, -1, "Number of rows:")
        self.number_rows=wx.StaticText(self, -1, "")
        self.number_cols_txt=wx.StaticText(self, -1, "Number of columns:")
        self.number_cols=wx.StaticText(self, -1, "")
        self.number_1s_txt=wx.StaticText(self, -1, "Number of 1s:")
        self.number_1s=wx.StaticText(self, -1, "")
        self.orig_nbfact_txt=wx.StaticText(self, -1, "Original number of factors:")
        self.orig_nbfact=wx.StaticText(self, -1, "")
        self.current_nbfact_txt=wx.StaticText(self, -1, "Current number of factors:")
        self.current_nbfact=wx.StaticText(self, -1, "")
        self.acc_error_txt=wx.StaticText(self, -1, "Accumulated error:")
        self.acc_error=wx.StaticText(self, -1, "")
        
        # Recommend
        self.Recommend_txt=wx.StaticText(self, -1, "Recommend Operation")
        self.operation_txt=wx.StaticText(self, -1, "Op:")
        self.operation=wx.StaticText(self, -1, "")
        self.operation_factors_txt=wx.StaticText(self, -1, "Factor (s):")
        self.operation_factors=wx.StaticText(self, -1, "")
        self.operation_error_txt=wx.StaticText(self, -1, "Error:")
        self.operation_error=wx.StaticText(self, -1, "")
        self.recommend_btn=wx.Button(self, 105, 'Recommend')
        self.visualize_btn=wx.Button(self, 106, 'Visualize')
        self.commit_btn=wx.Button(self, 107, 'Commit')
        self.skip_btn=wx.Button(self, 108, 'Skip')
        
        # Disable the Skip and Commit and Visualize buttons initially
        self.skip_btn.Disable()
        self.commit_btn.Disable()
        self.visualize_btn.Disable()
        
        #####
        # SIZERS
        #####
        # main sizer
        main_sizer=wx.BoxSizer(wx.HORIZONTAL)
        # sizer for the delete elements
        del_header_sizer=wx.BoxSizer(wx.VERTICAL)
        del_header_sizer.Add(self.Delete_txt, 0)
        
        del_sizer=wx.BoxSizer(wx.HORIZONTAL)
        del_sizer.Add(self.Delete_error_txt, 0, wx.RIGHT, border=30)
        del_sizer.Add(self.Delete_error, 0, wx.RIGHT, border=30)
        del_sizer.Add(self.Delete_btn, 0, wx.RIGHT, border=30)
        
        del_header_sizer.Add(del_sizer, 0)
        
        # sizers for the statistics part of the interface
        del_header_sizer.Add(self.statistics_txt, 0, wx.TOP, border=50)
        
        stats_left_sizer=wx.BoxSizer(wx.VERTICAL)
        stats_left_sizer.Add(self.number_rows_txt, 0)
        stats_left_sizer.Add(self.number_cols_txt, 0)
        stats_left_sizer.Add(self.number_1s_txt, 0)
        stats_left_sizer.Add(self.orig_nbfact_txt, 0)
        stats_left_sizer.Add(self.current_nbfact_txt, 0)
        stats_left_sizer.Add(self.acc_error_txt, 0)
        
        stats_right_sizer=wx.BoxSizer(wx.VERTICAL)
        stats_right_sizer.Add(self.number_rows, 0, wx.ALIGN_RIGHT)
        stats_right_sizer.Add(self.number_cols, 0, wx.ALIGN_RIGHT)
        stats_right_sizer.Add(self.number_1s, 0, wx.ALIGN_RIGHT)
        stats_right_sizer.Add(self.orig_nbfact, 0, wx.ALIGN_RIGHT)
        stats_right_sizer.Add(self.current_nbfact, 0, wx.ALIGN_RIGHT)
        stats_right_sizer.Add(self.acc_error, 0, wx.ALIGN_RIGHT)
        
        stats_left_right_sizer=wx.BoxSizer(wx.HORIZONTAL)
        stats_left_right_sizer.Add(stats_left_sizer)
        stats_left_right_sizer.Add(stats_right_sizer)
        
        del_header_sizer.Add(stats_left_right_sizer, 0)
        
        # sizer for the merge elements
        merg_header_sizer=wx.BoxSizer(wx.VERTICAL)
        merg_header_sizer.Add(self.Merge_txt, 0)
        
        merg_left_side_sizer=wx.BoxSizer(wx.VERTICAL)
        
        merg_OR_sizer=wx.BoxSizer(wx.HORIZONTAL)
        merg_OR_sizer.Add(self.Merge_OR_error_txt, 0, wx.RIGHT, border=30)
        merg_OR_sizer.Add(self.Merge_OR_error, 0, wx.RIGHT, border=70)
        
        merg_AND_sizer=wx.BoxSizer(wx.HORIZONTAL)
        merg_AND_sizer.Add(self.Merge_AND_error_txt, 0, wx.RIGHT, border=30)
        merg_AND_sizer.Add(self.Merge_AND_error, 0, wx.RIGHT, border=70)
        
        merg_left_side_sizer.Add(merg_OR_sizer, 0)
        merg_left_side_sizer.Add(merg_AND_sizer, 0)
        
        merg_btn_sizer=wx.BoxSizer(wx.VERTICAL)
        merg_btn_sizer.Add(self.Merge_Visualize_btn, 0)
        merg_btn_sizer.Add(self.Merge_OR_btn, 0)
        merg_btn_sizer.Add(self.Merge_AND_btn, 0)
        
        merg_inner_sizer=wx.BoxSizer(wx.HORIZONTAL)
        merg_inner_sizer.Add(merg_left_side_sizer, 0, wx.RIGHT, border=30)
        merg_inner_sizer.Add(merg_btn_sizer, 0, wx.RIGHT, border=30)
        
        merg_header_sizer.Add(merg_inner_sizer, wx.EXPAND)
        
        # sizer for the recommend operation
        recommend_header_sizer=wx.BoxSizer(wx.VERTICAL)
        recommend_header_sizer.Add(self.Recommend_txt, 0)
        
        left_side_sizer=wx.BoxSizer(wx.VERTICAL)
        
        operation_sizer=wx.BoxSizer(wx.HORIZONTAL)
        operation_sizer.Add(self.operation_txt, 0, wx.RIGHT, border=30)
        operation_sizer.Add(self.operation, 0, wx.RIGHT, border=30)
        
        operation_factors_sizer=wx.BoxSizer(wx.HORIZONTAL)
        operation_factors_sizer.Add(self.operation_factors_txt, 0, wx.RIGHT, border=30)
        operation_factors_sizer.Add(self.operation_factors, 0, wx.RIGHT, border=30)
        
        operation_error_sizer=wx.BoxSizer(wx.HORIZONTAL)
        operation_error_sizer.Add(self.operation_error_txt, 0, wx.RIGHT, border=30)
        operation_error_sizer.Add(self.operation_error, 0, wx.RIGHT, border=30)
        
        left_side_sizer.Add(operation_sizer, 0)
        left_side_sizer.Add(operation_factors_sizer, 0)
        left_side_sizer.Add(operation_error_sizer, 0)
        
        recommend_btns_sizer=wx.BoxSizer(wx.VERTICAL)
        recommend_btns_sizer.Add(self.recommend_btn, 0)
        recommend_btns_sizer.Add(self.visualize_btn, 0)
        recommend_btns_sizer.Add(self.commit_btn, 0)
        recommend_btns_sizer.Add(self.skip_btn, 0)
        
        recommend_inner_sizer=wx.BoxSizer(wx.HORIZONTAL)
        recommend_inner_sizer.Add(left_side_sizer, 0, wx.RIGHT, border=30)
        recommend_inner_sizer.Add(recommend_btns_sizer, 0, wx.RIGHT, border=30)
        
        recommend_header_sizer.Add(recommend_inner_sizer, wx.EXPAND)

        # set Boxsizer with all sizers
        main_sizer.Add(del_header_sizer, 1, wx.EXPAND)
        main_sizer.Add(merg_header_sizer, 1, wx.EXPAND)
        main_sizer.Add(recommend_header_sizer, 1, wx.EXPAND)
        
        self.SetSizer(main_sizer)
        
        
####
# grid table for the top panel of the first tab (tab1)wx.RIGHT, border=10
####
class TopPanelTable(gridlib.PyGridTableBase):
    
    
    def __init__(self, data, rowLabels=None, colLabels=None):
        gridlib.PyGridTableBase.__init__(self)
        
        # initialize the arguments
        self.data=data
        self.rowLabels=rowLabels
        self.colLabels=colLabels
        
        # length of previous data to be sure the table changed/ the user clicked the open button again
        self.previous_rows=0
    
    
    def GetNumberRows(self):
        return len(self.data)
    
    def GetNumberCols(self):
        return len(self.colLabels)
    
    def GetColLabelValue(self,col):
        if self.colLabels:
            return self.colLabels[col]
        
    def GetRowLabelValue(self,row):
        if self.rowLabels:
            return self.rowLabels[row]
        
    def isEmptyCell(self, row, col):
        # returns True if the cell is empty
        if self.GetValue(row, col)!="":
            return False
        else:
            return True
        
    def GetValue(self, row, col):
        if self.data:
            return self.data[row][col]
        else:
            return ''
    
    def SetValue(self, row, col, value):
        # convert data to list
        data=list(self.data)
        # get the row for which we want to set the value
        table_row=list(data[row])
        # set the value
        table_row[col]=value
        # copy back to data
        data[row]=tuple(table_row)
        # convert data back to tuple
        self.data=tuple(data)
        
    def insertData(self, originaldata):
        # set data
        self.data=originaldata
        
        # number of records to append
        number_rows_to_insert=len(originaldata)
        
        # set row labels
        self.rowLabels=tuple(str(i) for i in range(_rows+1, number_rows_to_insert+1))
        
        # first delete notify the table to delete all previous rows
        self.GetView().ProcessTableMessage(gridlib.GridTableMessage(self,gridlib.GRIDTABLE_NOTIFY_ROWS_DELETED, 0, self.previous_rows))
        
        # notify the grid table
        self.GetView().ProcessTableMessage(gridlib.GridTableMessage(self,gridlib.GRIDTABLE_NOTIFY_ROWS_INSERTED, 0, number_rows_to_insert))
       
        # update the previous number of rows
        self.previous_rows=self.GetNumberRows()
        
    def RemoveData(self,rowList):
        # update the data
        self.data=tuple(x for x in self.data if self.data.index(x) not in rowList)
        
        # set row labels
        self.rowLabels=tuple(str(i) for i in range(_rows+1, self.GetNumberRows()+1))
         
        # notify the grid table that we have deleted some rows
        self.GetView().ProcessTableMessage(gridlib.GridTableMessage(self,gridlib.GRIDTABLE_NOTIFY_ROWS_DELETED, len(self.data), len(rowList)))
        
    def AddData(self, newdata):
        
        # set data
        self.data=self.data + newdata
        
        # number of records to append
        number_to_append=len(newdata)
        
        # set row labels
        self.rowLabels=tuple(str(i) for i in range(_rows+1, self.GetNumberRows()+1))
        
        # notify the grid table
        self.GetView().ProcessTableMessage(gridlib.GridTableMessage(self,gridlib.GRIDTABLE_NOTIFY_ROWS_APPENDED,number_to_append))
        
    def SortColumn(self, col):
        
        if col in [0, 1, 2, 3]:
            
            # get the column values as a list
            selected_col= [self.data[row][col] for row in range(len(self.data))]
            
        else:
            
            selected_col= [int(self.data[row][col]) for row in range(len(self.data))]
        
        # sort and return the indices of sorted list
        # enumerate (selected_col gives a list of tuples (index, value)
        # which we sort using sorted
        index_value_tuples=sorted(enumerate(selected_col), key=lambda x:x[1])
        
        indx_sorted=[i[0] for i in index_value_tuples]
        
        # sorted data
        self.data=[self.data[indx] for indx in indx_sorted]
        
        # sort the factors as well so that the information corresponds
        dt.factoritems_=[dt.factoritems_[i] for i in indx_sorted]
    
    def Refresh(self):
        # this method refreshes the entire grid display or causes it to redraw after sorting
        self.GetView().ProcessTableMessage(gridlib.GridTableMessage(self, gridlib.GRIDTABLE_REQUEST_VIEW_GET_VALUES))
        
####
# DATA GENERATION FUNCTIONS
# functions that generated the data needed to plot on the table grid
####

# this function generates the data at the beginning as the grid is loaded first
def TableDataGenerate(factors, closeditems, transactions, scores):

    # column labels
    # this is fixed by the design
    colLabels=("Column names", "Number of items", "Row names", "Number of transactions", "Area", "Unique area", "Error if deleted")
    # just incremental from 1
    rowLabels=tuple(str(i) for i in range(1, len(factors)+1))

    
    # the data is obtained by querying the various data structures which contain the needed information
    data=()
    
    if factors:
        for factor in factors:
            # get the items that make up the factor and the transactions in which the factor is present
            items_factor=closeditems[factor]
            trans_factor=transactions[factor]
            
            # get the corresponding column names and row names by calling the function get_col_row_names
            # set some variables which are needed for the Visualization_functions module to run successfully
            # change the path to point to the input
            global path_input_
            global inp
            global trans_1s # each transaction and 1 where an item is in the transaction and 0 otherwise
            global cols_1s
            
            path_input_=path_input_f+"/input"
            
            # call the Exact_bool_factorization module to get some details of the input data
            input_data_details=ebf.input_read(path_input_)
            
            # update global variables
            inp= input_data_details[0]
            trans_1s=input_data_details[1]
            dt.indices_1s_=input_data_details[2]
            dt.number_columns_=input_data_details[3]
            dt.number_rows_=input_data_details[4]
            
            # get the number of 1s in each column from the indices
            cols_indices=dt.indices_1s_[:,1]
            
            # transform the array into a dictionary of unique column indices with the number of 1s as values
            unique, counts=np.unique(cols_indices, return_counts=True)
            cols_1s=dict(zip(unique, counts))
            
            # call to get the names of columns and rows
            vs.get_set_values(path_input_, inp)
            
            # now get the column and row names 
            rows_names, cols_names=vs.get_col_row_names()
            
            # select the row and col names of the factor in question
            rows_names_factor=[x for x in rows_names if rows_names.index(x) in trans_factor]
            # convert to string
            rows_names_factor_str=', '.join(str(e) for e in rows_names_factor)
            
            cols_names_factor=[x for x in cols_names if cols_names.index(x) in items_factor]
            # convert to string
            cols_names_factor_str=', '.join(str(e) for e in cols_names_factor)
            
            # number of items
            nb_items=len(cols_names_factor)
            
            # number of transactions
            nb_trans=len(rows_names_factor)
            
            # the area is the total amount of area covered by the factor which is the number of items times the number of transactions
            area= str(len(items_factor)*len(trans_factor))
            
            # the unique area is the number of 1s covered alone by this factor
            unique_area=str(scores[factor])
             
            # error if deleted is theb same as the unique area
            error=str(unique_area)
            
            # here is a row of the grid table
            # after going through the whole for loop we have all the rows corresponding to the number of factors 
            data_row=(cols_names_factor_str, nb_items, rows_names_factor_str, nb_trans, area, unique_area, error,)
            
            data=data + (data_row,)
        
    
    else:
        data=()
    
    return colLabels, rowLabels, data

# this function generates the column name and row names given the factor every time a new factor needs to be added due to a merge operation

def row_col_factor(closeditems, transactions, factor_index):
            
            global path_input_
            global inp
            
            # get the items that make up the factor and the transactions in which the factor is present
            items_factor=closeditems[factor_index]
            trans_factor=transactions[factor_index]
            
            # call to get the names of columns and rows
            vs.get_set_values(path_input_, inp)
            
            # now get the column and row names 
            rows_names, cols_names=vs.get_col_row_names()
            
            # select the row and col names of the factor in question
            rows_names_factor=[x for x in rows_names if rows_names.index(x) in trans_factor]
            # convert to string
            rows_names_factor_str=','.join(str(e) for e in rows_names_factor)
            
            cols_names_factor=[x for x in cols_names if cols_names.index(x) in items_factor]
            # convert to string
            cols_names_factor_str=','.join(str(e) for e in cols_names_factor)
            
            # number of items
            nb_items=len(cols_names_factor)
            
            # number of transactions
            nb_trans=len(rows_names_factor)
            
            
            return rows_names_factor_str, nb_trans, cols_names_factor_str, nb_items
        
        
# the function below generates the data for the rows and columns tabs

def Rows_Cols_TableDataGenerate():
    
    # rows
    # column labels
    # this is fixed by the design
    row_colLabels=("Row Number", "Number of 1s")
#     # just incremental from 1
#     row_rowLabels=tuple(str(i) for i in range(1, len(inp)+1))
    
    # columns
    # column labels
    # this is fixed by the design
    col_colLabels=("Column Number", "Number of 1s")
#     # just incremental from 1
#     col_rowLabels=tuple(str(i) for i in range(1, len(inp[0])+1))
    
    # generate the data for the rows and columns tabs
    rows_data=()
    cols_data=()
        
    if dt.factoritems_:
        
        for row_number in range(len(inp)):
            row_data=(row_number, len(trans_1s[row_number]))
        
        rows_data=rows_data + (row_data,)
        
        for col_number in range(len(inp[0])):
            col_data=(col_number, len(cols_1s[col_number]))
            
        cols_data=cols_data + (col_data,)
    
#     return row_colLabels, row_rowLabels, rows_data, col_colLabels, col_rowLabels, cols_data
    return row_colLabels, rows_data, col_colLabels, cols_data
    

####
# NOTEBOOK TAB CLASSES
# panel for the second tab (tab2)
# the tab displays the rows and some statistics about them
####

class tab_rows(wx.Panel):
     
    def __init__(self, parent):
        wx.Panel.__init__(self, parent)

        # some static text
        t=wx.StaticText(self, -1, "Here are displayed the rows and some statistics about them", (20,20))
        
####
# panel for the third tab (tab3) 
# the tab displays the columns and some statistics about them
####  
 
class tab_cols(wx.Panel):
     
    def __init__(self, parent):
        wx.Panel.__init__(self, parent)

        # some static text
        t=wx.StaticText(self, -1, "Here are displayed the columns and some statistics about them", (20,20))
          
####
# MAIN CLASS (THE FRAME)
####    
class interactive_BMF(wx.Frame):
    
    def __init__(self):
        
        # initialize frame
        default = wx.MINIMIZE_BOX | wx.MAXIMIZE_BOX | wx.RESIZE_BORDER | wx.SYSTEM_MENU | wx.CAPTION | wx.CLOSE_BOX | wx.CLIP_CHILDREN
        wx.Frame.__init__(self, None, -1, title="Interactive BMF System", size=(700,700),style=default)
        self.Center()
        
        # main  panel
        self.Main_panel=wx.Panel(self)
        
        # create a panel and notebook (tabs holder)
        self.nb=wx.Notebook(self.Main_panel)
        
        
        # panel
        self.panel = wx.Panel(self.nb)
        # splitter for the two panels in the first tab
        self.splitter=wx.SplitterWindow(self.panel)
        self.TPanel=TopPanel(self.splitter)                             
        self.LPanel=LowPanel(self.splitter)
        
        # split the window
        self.splitter.SplitHorizontally(self.TPanel, self.LPanel, 500)
        
        #####
        # GRID
        #####
        self.grid = gridlib.Grid(self.TPanel)
        self.grid.EnableEditing(False)
        # data and headers
        colLabels, rowLabels, data= TableDataGenerate(dt.factoritems_, dt.clitems_, dt.trans_sets_, dt.scores_factors_)
        self._data=TopPanelTable(data, rowLabels, colLabels)
        self.grid.SetTable(self._data)
        
        # make the grid to fill the whole panel
        self.grid.AutoSizeRows()
        self.grid.AutoSizeColumns()
        
        
        # handlers for the selection of cell or cells
        self.grid.Bind(gridlib.EVT_GRID_SELECT_CELL, self.OnSingleSelect)
#         self.grid.Bind(gridlib.EVT_GRID_LABEL_LEFT_CLICK, self.OnSingleSelect_LeftClick)
        self.grid.Bind(gridlib.EVT_GRID_RANGE_SELECT, self.OnDragSelection)
        # on double click the local view is displayed
        self.grid.Bind(gridlib.EVT_GRID_LABEL_LEFT_DCLICK, self.OnDoubleSelect_LeftClick)
        
        # click event on the column label for the sorting
        self.grid.Bind(gridlib.EVT_GRID_LABEL_RIGHT_CLICK, self.OnLabelRightClicked)
        
        # create the other tabs for rows and columns
        tab2=tab_rows(self.nb)
        tab3=tab_cols(self.nb)
         
        # Add the windows to tabs and name them
        self.nb.AddPage(self.panel, "Factors")
        self.nb.AddPage(tab2, "Rows")
        self.nb.AddPage(tab3, "Columns")
        

        # sizer for the grid
        sizer_grid=wx.BoxSizer()
        sizer_grid.Add(self.grid, 1, wx.EXPAND|wx.ALL)
        self.TPanel.SetSizer(sizer_grid)
        
        
        # sizer for the panel which is the first page of the notebook
        sizer_panel=wx.BoxSizer()
        sizer_panel.Add(self.splitter, 1, wx.EXPAND)
        self.panel.SetSizer(sizer_panel)
        
        # put the notebook in a sizer for the panel to manage the layout
        sizer=wx.BoxSizer()
        sizer.Add(self.nb, 1, wx.EXPAND)
        self.Main_panel.SetSizer(sizer)
    
        
        ####
        # FRAME
        ####                  
        # create status bar
        status_bar=self.CreateStatusBar()
        
        # create menu bar
        menu_bar=wx.MenuBar()
        first=wx.Menu()
        second=wx.Menu()
        fileOpenMenuItem=first.Append(101,"Open","Load cPickled files from disk")
        fileExitMenuItem=first.Append(102,"Exit","Exit program")
        helpmenuItem=second.Append(103,"Help", "Get Help!")
        menu_bar.Append(first,"File")
        menu_bar.Append(second,"Help")
        self.SetMenuBar(menu_bar)
        
        # bind an event to the menu item
        self.Bind(wx.EVT_MENU, self.OnOpen, fileOpenMenuItem)
        self.Bind(wx.EVT_MENU, self.closewindow, fileExitMenuItem)
        # closing button
        self.Bind(wx.EVT_CLOSE, self.closewindow)
#         self.Bind(wx.EVT_MENU, self.OnHelp, helpmenuItem)
        
        # handlers of buttons actually on the lower panel
        self.Bind(wx.EVT_BUTTON, self.OnDelete, self.LPanel.Delete_btn)
        self.Bind(wx.EVT_BUTTON, self.OnMergeOR, self.LPanel.Merge_OR_btn)
        self.Bind(wx.EVT_BUTTON, self.OnMergeAND, self.LPanel.Merge_AND_btn)
        self.Bind(wx.EVT_BUTTON, self.OnRecommend, self.LPanel.recommend_btn)
        self.Bind(wx.EVT_BUTTON, self.OnSkip, self.LPanel.skip_btn)
        self.Bind(wx.EVT_BUTTON, self.OnVisualize, self.LPanel.visualize_btn)
        self.Bind(wx.EVT_BUTTON, self.OnMergeVisualize, self.LPanel.Merge_Visualize_btn)
        self.Bind(wx.EVT_BUTTON, self.OnCommit, self.LPanel.commit_btn)
        #### 
        # THREAD RELATED CODE
        ####
        
        # set up the event handler for any worker thread results
        EVT_RESULT(self, self.OnDisplayWorkerResult)
         
        # and indicate we don't have a worker yet
        self.worker=None
        
        
    ####
    # FUNCTIONS/EVENT HANDLERS
    # Method to open the cPickled files and load them into the global variables so they are accessible to all classes through the important_data module
    ####	
    def OnOpen(self,event):
    
        global path_input_f
        
        # empty all static texts which may be not be empty already
        self.LPanel.Delete_error.SetLabel("")
        self.LPanel.Merge_OR_error.SetLabel("")
        self.LPanel.Merge_AND_error.SetLabel("")
        self.LPanel.operation.SetLabel("")
        self.LPanel.operation_error.SetLabel("")
       
        
        dlg=wx.DirDialog(self, message="Select the cPickled folder created in the same directory as your input data...")
        
        if dlg.ShowModal()==wx.ID_OK:
            # go recursively in the folder and return all cPickle files
            path=dlg.GetPath()
            
            # update global variable
            path_input_f=os.path.dirname(path)
            
            for paths_f in os.walk(path):
                # get the path of the cPickled folder
                path_cPickled=paths_f[0]
                
                try:
                    bkp_closed_f=path_cPickled+'/cPickled_closed_sets'
                    with open(bkp_closed_f, "r") as fp:
                        dt.clitems_ = cPickle.load(fp)
                        
                    bkp_trans_f=path_cPickled+'/cPickled_trans_sets'
                    with open(bkp_trans_f, "r") as fp:
                        dt.trans_sets_ = cPickle.load(fp)
                        
                    bkp_cover_f=path_cPickled+'/cPickled_cover'
                    with open(bkp_cover_f, "r") as fp:
                        dt.cover_ = cPickle.load(fp)
                        
                    bkp_factors_f=path_cPickled+'/cPickled_factor_items'
                    with open(bkp_factors_f, "r") as fp:
                        dt.factoritems_ = cPickle.load(fp)
                    
                    bkp_number_1s_f=path_cPickled+'/cPickled_number_1s'
                    with open(bkp_number_1s_f, "r") as fp:
                        dt.number_1s_ = cPickle.load(fp)
                    
                    bkp_reduced_cover_f=path_cPickled+'/cPickled_reduced_indx_cover'
                    with open(bkp_reduced_cover_f, "r") as fp:
                        dt.reduced_index_cover_ = cPickle.load(fp)
                    
                    bkp_scores_f=path_cPickled+'/cPickled_scores'
                    with open(bkp_scores_f, "r") as fp:
                        dt.scores_factors_=cPickle.load(fp)
                    
                    bkp_error_f=path_cPickled+'/cPickled_error'
                    with open(bkp_error_f, "r") as fp:
                        dt.acc_error_=cPickle.load(fp)
                    
                    # sort the factors in ascending order of their unique area (score) before generating the data
                    # to sort the factors we need to sort their scores
                    sorted_scores=sorted(dt.scores_factors_.items(), key= lambda item: item[1])
                    
                    # now we sort the factors
                    dt.factoritems_=[x for (x, y) in sorted_scores]
                   
                    # generate the data and it to the grid
                    colLabels, rowLabels, data= TableDataGenerate(dt.factoritems_, dt.clitems_, dt.trans_sets_, dt.scores_factors_)
                    
                    # then we write the new data
                    self.grid.GetTable().insertData(data)

                 
                    # also update the statistics part of the interface
                    self.LPanel.number_rows.SetLabel(str(dt.number_rows_))
                    self.LPanel.number_cols.SetLabel(str(dt.number_columns_))
                    self.LPanel.number_1s.SetLabel(str(len(dt.indices_1s_)))
                    self.LPanel.orig_nbfact.SetLabel(str(len(dt.factoritems_)))
                    self.LPanel.current_nbfact.SetLabel(str(len(dt.factoritems_)))
                    self.LPanel.acc_error.SetLabel(str(dt.acc_error_))
                    
                    # break out of the for loop
                    break

                        
                except:
                    # in case some file is missing 
                    msg_dlg=wx.MessageDialog(None, "Make sure you have made no changes to the cPickled folder and that all the eight (8) files from exact factorization are present.", "ERROR", wx.OK|wx.ICON_ERROR)
                    
                    if msg_dlg.ShowModal()==wx.ID_OK:
                        msg_dlg.Destroy()
                    else:
                        msg_dlg.Destroy()
        else:
            dlg.Destroy()
        
    
    #####
    # EVENT HANDLERS
    ####
    def OnLabelRightClicked(self, event):
        # get the selected column
        col= event.GetCol()
        
        # do a popup menu and with a sort column menu item 
        # and the functionality to sort
        # display a popup menu when a column label is right clicked
        x= self.grid.GetColSize(col)/2
        menu=wx.Menu()
        sortID=wx.NewId()
        
        xo, yo= event.GetPosition()
#         self.grid.Selectcol(col)
        
        cols=self.grid.GetSelectedCols()
        self.Refresh()
        
        menu.Append(sortID, "Sort Column")
        
        def sort(event, self=self, col=col):
            self.grid.GetSelectedCols()
            self.grid.GetTable().SortColumn(col)
            
            # refresh the grid
            self.grid.GetTable().Refresh()
        
        self.grid.Bind(wx.EVT_MENU, sort, id=sortID)
        
        self.grid.PopupMenu(menu)
        menu.Destroy()
        return
        
    
    # display the local view when the user doubles clicks the label with the left mouse click
    def OnDoubleSelect_LeftClick(self, event):

        # call the function to do the plotting (local view)
        self.PlotLocal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                            , dt.scores_factors_, int(event.GetRow()), "", "")

    
    
    def OnSingleSelect(self, event):
        
        global delete_list
        global curr_row
        global merge_pair
        global count
        
        count+=1
        # set the merge pair variable to empty since the user just clicked another factor
        merge_pair=[]
        
        # make the other labels empty
        self.LPanel.Merge_OR_error.SetLabel("")
        self.LPanel.Merge_AND_error.SetLabel("")
        
        # first clean the delete static text
        self.LPanel.Delete_error.SetLabel("")
        
        # get the selection of a single cell by clicking or moving the selection with the arrow keys
        # display the selection in the lower panel for the user to decide if he deletes the factor or not
        row_nb=int(event.GetRow())
        self.LPanel.Delete_error.SetLabel(str(dt.scores_factors_[dt.factoritems_[row_nb]]))
        
        # update the global view every time a new cell is clicked
        if count==1:
            self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_\
																,dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                            					, dt.scores_factors_,int(event.GetRow()) , "", "")
        else:
            
            self.inst_plot_visual.Destroy()
            
            self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_\
                                                                ,dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                                                , dt.scores_factors_,int(event.GetRow()) , "", "")
        
        # update current row
        curr_row=event.GetRow()
        
        
        event.Skip()

    def OnDragSelection(self, event):
        
        global delete_list
        global curr_row # currently selected row
        global curr_rows # currently selected rows
        global merge_pair
        
        selection=self.grid.GetSelectedCells()
        
        # list of selected rows
        selected_rows=[]
        
       
        if selection:
            
            # compute the error of deleting all selected and display on the low panel
            total_error=0
            
                            
            for x in selection:
                if x[0] not in selected_rows:
                    total_error+=dt.scores_factors_[dt.factoritems_[int(x[0])]]
                    selected_rows.append(x[0])
                     
                    # add the chosen factor to the factors to delete
                    curr_rows=selected_rows
            
            # check if the drag is over two factors since the user might want to merge the two factors
            if len(selection)==2:
                
                # set some variables in the Constraints_new module
                cnstr.get_set_values(inp, dt.indices_1s_, dt.number_columns_, dt.number_rows_)
                
                # call the error_OR_AND_optimized function from the Constraints_new module to get the errors of merge with OR or AND
                similar_pair=(selection[0][0], selection[1][0],)
                
                self.errors_OR_AND_optimal= cnstr.error_OR_AND_optimized(similar_pair, dt.clitems_, dt.trans_sets_, dt.factoritems_)
                
                # retrieve the errors
                error_OR= self.errors_OR_AND_optimal[0]
                error_AND= self.errors_OR_AND_optimal[1]
                # update the labels of the static texts
                self.LPanel.Merge_OR_error.SetLabel(str(error_OR))
                self.LPanel.Merge_AND_error.SetLabel(str(error_AND))
                
                # update the pair to merge
                merge_pair=similar_pair
                
                
            else:
                # empty the labels of the merge static texts since we can merge only two factors
                self.LPanel.Merge_OR_error.SetLabel("")
                self.LPanel.Merge_AND_error.SetLabel("")
                
            # set the label of the delete error static text
            self.LPanel.Delete_error.SetLabel(str(total_error))
                
                
    def OnDelete(self, event):
        
        # global variables
        global delete_list
        global curr_row
        global curr_rows
         
        # get the current selected rows
        if curr_row!="":
            delete_list.append(int(curr_row))
         
        for x in curr_rows:
            delete_list.append(x)
       
        # convert the delete list to a set to get only unique values
        delete_list_set=set(delete_list)
        
        # delete the data provided in the arguments
        self.grid.GetTable().RemoveData(list(delete_list_set))
        
        # update the global variables to reflect the changes
        # update the factors
        new_factors=[x for x in dt.factoritems_ if dt.factoritems_.index(x) not in delete_list]
                
        # factors in delete list
        delete_list_factors=[dt.factoritems_[x] for x in delete_list]
        
        # update the scores
        # by deleting some factors, the number of ones covered alone by some other factors may increase
        # so we look through the reduced_cover and update the scores accordingly
        for k in dt.reduced_index_cover_.keys():
            # we only consider the indices which are covered only by 2 factors since these will be covered by only 1 if the other is deleted
            if len(dt.reduced_index_cover_[k])==2:
                  
                factor_to_update=list(set(dt.reduced_index_cover_[k]).difference(set(delete_list_factors)))
                  
                # if the difference is not empty then it contains the factor whose score should be incremented by 1
                if len(factor_to_update)==1:
                    if factor_to_update[0] in dt.scores_factors_.keys():
                        
                        dt.scores_factors_[factor_to_update[0]]+=1
                      
                        # update the unique area and error of this factor from the data in the grid
                        self.grid.GetTable().SetValue(new_factors.index(factor_to_update[0]), 3, dt.scores_factors_[factor_to_update[0]])
                        self.grid.GetTable().SetValue(new_factors.index(factor_to_update[0]), 4, dt.scores_factors_[factor_to_update[0]]) 
               
        # update the accumulated error
        for el in delete_list_factors:
            dt.acc_error_+=dt.scores_factors_[el]
             
                  
        # remove the factors deleted from the scores
        dt.scores_factors_={i: dt.scores_factors_[i] for i in dt.scores_factors_.keys() if i not in delete_list_factors}
        
        # finally update the factors
        dt.factoritems_=new_factors
        
        # update the statistics portion of the interface
        self.LPanel.current_nbfact.SetLabel(str(len(dt.factoritems_)))
        self.LPanel.acc_error.SetLabel(str(dt.acc_error_))
        
        # set the curr_row, curr_rows, and delete_list by setting to empty for the next delete
        curr_row=""
        curr_rows=[]
        delete_list=[]
        
        # reset the labels for Delete and Merge to empty
        self.LPanel.Delete_error.SetLabel("")
        self.LPanel.Merge_OR_error.SetLabel("")
        self.LPanel.Merge_AND_error.SetLabel("")
        
        # reset the recommendation labels to empty
        self.LPanel.operation.SetLabel("")
        self.LPanel.operation_factors.SetLabel("")
        self.LPanel.operation_error.SetLabel("")
        
        # update global view
        # destroy previous window
        self.inst_plot_visual.Destroy()
        # plot global view
        self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                        , dt.scores_factors_, 0, "d", "")
        
    
    def OnMergeOR(self, event):

    	global merge_pair
          
         
        if self.LPanel.Merge_OR_error.LabelText=="":
              
            msg_error=wx.MessageDialog(self, "Please select a pair to merge", "Error", wx.OK|wx.ICON_ERROR)
              
            if msg_error.ShowModal()==wx.ID_OK:
                msg_error.Destroy()
        else:
                  
            # first we delete the factors in the merge pair from the grid
            self.grid.GetTable().RemoveData(list(merge_pair))
              
            # update the global data variables
            dt.clitems_=self.errors_OR_AND_optimal[2]
            dt.trans_sets_=self.errors_OR_AND_optimal[3]
              
            # do not update the factors yet
            new_factors=self.errors_OR_AND_optimal[4]
              
            # then update the number of 1s    
            dt.number_1s_[new_factors[-1]]=len(dt.trans_sets_[-1])*len(dt.clitems_[-1])
              
            # we need to update the scores
            # first get the number of 1s that each factor in the pair to merge covers alone
            merge_pair_unique_area=0
              
            for x in merge_pair:
                merge_pair_unique_area+=dt.scores_factors_[dt.factoritems_[x]]
                  
                  
            # delete first the keys and values corresponding to the pair to merge
            for x in merge_pair:
                del dt.scores_factors_[dt.factoritems_[x]]
               
            # update the factoritems
            dt.factoritems_=new_factors
                  
            # then update the scores
            # if the new factor covers no one alone there is no need to update the scores
            if merge_pair_unique_area!=0:
                  
                dt.scores_factors_[dt.factoritems_[-1]]=merge_pair_unique_area
              
                # generate new data
                # get the row and column names of the factor resulting from the merge
                rows_names_merged, nb_trans, col_names_merged, nb_items=row_col_factor(dt.clitems_, dt.trans_sets_, dt.factoritems_[-1]) 
                  
                # new row to add to the grid
                # we only add the new data if the score is not 0 (if the score is 0 there is no need to add that factor)
                data=()
                new_data_row=(col_names_merged, nb_items, rows_names_merged, nb_trans, dt.number_1s_[dt.factoritems_[-1]], dt.scores_factors_[dt.factoritems_[-1]], dt.scores_factors_[dt.factoritems_[-1]],)
                data = data + (new_data_row,)
                  
                # call the AddData method of the pygridTablebase class to update the grid
                self.grid.GetTable().AddData(data)
              
            else:
                  
                # since the unique area of the new factor is zero, we remove it from the list of factors
                dt.factoritems_=dt.factoritems_[:-1]
              
            # update the accumulated error
            dt.acc_error_+=self.errors_OR_AND_optimal[0]
              
            # update the statistics portion of the interface
            self.LPanel.current_nbfact.SetLabel(str(len(dt.factoritems_)))
            self.LPanel.acc_error.SetLabel(str(dt.acc_error_))
              
            # set the curr_row, curr_rows, delete_list, and merge_pair by setting to empty for the next delete
            curr_row=""
            curr_rows=[]
            delete_list=[]
            merge_pair=[]
              
            # reset the labels for Delete and Merge to empty
            self.LPanel.Delete_error.SetLabel("")
            self.LPanel.Merge_OR_error.SetLabel("")
            self.LPanel.Merge_AND_error.SetLabel("")
              
            # reset the recommendation labels to empty
            self.LPanel.operation.SetLabel("")
            self.LPanel.operation_factors.SetLabel("")
            self.LPanel.operation_error.SetLabel("")
                  
            # update global view
            # destroy previous window
            self.inst_plot_visual.Destroy()
            # plot global view
            self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                            , dt.scores_factors_, 0, "d", "")
                  
    def OnMergeAND(self, event):
        
        global merge_pair
        
        if self.LPanel.Merge_AND_error.LabelText=="":
            
            msg_error=wx.MessageDialog(self, "Please select a pair to merge", "Error", wx.OK|wx.ICON_ERROR)
            
            if msg_error.ShowModal()==wx.ID_OK:
                msg_error.Destroy()
        else:
 
            # first we delete the factors in the merge pair from the grid
            self.grid.GetTable().RemoveData(list(merge_pair))
            
            # update the global data variables
            dt.clitems_=self.errors_OR_AND_optimal[5]
            dt.trans_sets_=self.errors_OR_AND_optimal[6]
            new_factors=self.errors_OR_AND_optimal[7]
            
            # we also need to update the number of 1s
            # delete first the keys and values corresponding to the pair to merge
            for x in merge_pair:
                del dt.number_1s_[dt.factoritems_[x]]
            
            # then update the number of 1s    
            dt.number_1s_[dt.factoritems_[-1]]=len(dt.trans_sets_[-1])*len(dt.clitems_[-1])
            
            # we need to update the scores
            # first get the indices that each factor in the merge pair covers
            indx_factor_1=[(x,y) for x in dt.trans_sets_[dt.factoritems_[merge_pair[0]]] for y in dt.clitems_[dt.factoritems_[merge_pair[0]]]]
            indx_factor_2=[(x,y) for x in dt.trans_sets_[dt.factoritems_[merge_pair[1]]] for y in dt.clitems_[dt.factoritems_[merge_pair[1]]]]
            
            # the length of the intersection of the two sets of indices above will give us the number of 1s that the merged factor covers
            merge_pair_unique_area=len(set(indx_factor_1).intersection(set(indx_factor_2)))
            
            # delete first the keys and values corresponding to the pair to merge
            for x in merge_pair:
                del dt.scores_factors_[dt.factoritems_[x]]
            
            # factors in merge pair
            merge_pair_factors=[dt.factoritems_[x] for x in merge_pair]
            
            # update the scores
            # by deleting some factors, the number of ones covered alone by some other factors may increase
            # so we look through the reduced_cover and update the scores accordingly
            for k in dt.reduced_index_cover_.keys():
                # we only consider the indices which are covered only by 2 factors since these will be covered by only 1 if the other is deleted
                if len(dt.reduced_index_cover_[k])==2:
                      
                    factor_to_update=list(set(dt.reduced_index_cover_[k]).difference(set(merge_pair_factors)))
                      
                    # if the difference is not empty then it contains the factor whose score should be incremented by 1
                    if len(factor_to_update)==1:
                        if factor_to_update[0] in dt.scores_factors_.keys():
                            
                            dt.scores_factors_[factor_to_update[0]]+=1
                          
                            # update the unique area and error of this factor from the data in the grid
                            self.grid.GetTable().SetValue(new_factors.index(factor_to_update[0]), 5, dt.scores_factors_[factor_to_update[0]])
                            self.grid.GetTable().SetValue(new_factors.index(factor_to_update[0]), 6, dt.scores_factors_[factor_to_update[0]])
            
            # finally update the factors
            dt.factoritems_=new_factors
                
            # we add the new factor and its score if only the area covered by the new factor is not zero 
            if merge_pair_unique_area!=0:   
                 
                # then update the scores
                dt.scores_factors_[dt.factoritems_[-1]]=merge_pair_unique_area
            
                # generate new data
                # get the row and column names of the factor resulting from the merge
                rows_names_merged, nb_trans, col_names_merged, nb_items=row_col_factor(dt.clitems_, dt.trans_sets_, dt.factoritems_[-1])  
                
                # new row to add to the grid
                data=()
                new_data_row=(col_names_merged, nb_items, rows_names_merged, nb_trans, dt.number_1s_[dt.factoritems_[-1]], dt.scores_factors_[dt.factoritems_[-1]], dt.scores_factors_[dt.factoritems_[-1]],)
                data = data + (new_data_row,)
                
                # call the AddData method of the pygridTablebase class to update the grid
                self.grid.GetTable().AddData(data)
                
            
            else:
                
                # since the unique area of the new factor is zero, we remove it from the list of factors
                dt.factoritems_=dt.factoritems_[:-1]
                
            
            # update the accumulated error
            dt.acc_error_+=self.errors_OR_AND_optimal[1]
            
            # update the statistics portion of the interface
            self.LPanel.current_nbfact.SetLabel(str(len(dt.factoritems_)))
            self.LPanel.acc_error.SetLabel(str(dt.acc_error_))
            
            # set the curr_row, curr_rows, delete_list, and merge_pair by setting to empty for the next delete
            curr_row=""
            curr_rows=[]
            delete_list=[]
            merge_pair=[]
            
            # reset the labels for Delete and Merge to empty
            self.LPanel.Delete_error.SetLabel("")
            self.LPanel.Merge_OR_error.SetLabel("")
            self.LPanel.Merge_AND_error.SetLabel("")
            # reset the recommendation labels to empty
            self.LPanel.operation.SetLabel("")
            self.LPanel.operation_factors.SetLabel("")
            self.LPanel.operation_error.SetLabel("")
                
            # update global view
            # destroy previous window
            self.inst_plot_visual.Destroy()
            # plot global view
            self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                            , dt.scores_factors_, 0, "d", "")
           
    def OnRecommend(self, event):
        
        global skip_count
        
        # empty the variables which do the book keeping when the user skips the recommended operation
        dt.error_index_delete_={}
        dt.error_similar_pair_={}
        # also reset the skip count to zero
        skip_count=0
        
        if self.grid.GetTable().isEmptyCell(0,0):
           
            msg_error=wx.MessageDialog(self, "Load data before clicking recommend", "Error", wx.OK|wx.ICON_ERROR)
            
            if msg_error.ShowModal()==wx.ID_OK:
                msg_error.Destroy()
        
        else:
            
            # triggers the worker thread unless it's already busy
            if not self.worker:
                # trigger worker by passing it the data it needs to carry out task
                self.worker= RecommendThread(self, inp, dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                                  , dt.scores_factors_)
                self.worker.start()

                    
    ####
    # THREAD RELATED EVENT HANDLER
    ####
    def OnDisplayWorkerResult(self, event):
         
        # after the worker is done with the computation display the recommended operation in lower panel
        if event.data!=True:
            # update global variables
            # update some more global variables
            dt.similar_pairs_=event.data[8] # similar pairs
            dt.ordered_factors_items_=event.data[9] # factors sorted in ascending order of their unique area
            
            # set the label for the operation recommended
            if event.data[0]=="d":
                self.LPanel.operation.SetLabel("Delete")
                self.LPanel.operation_factors.SetLabel(str(event.data[2]+1))
                self.LPanel.operation_error.SetLabel(str(event.data[1]))
                
                # update the user choice
                dt.user_choose_=event.data[0]
                
                # update the index recommended to merge
                dt.factorindx_=event.data[2]
                 
                # update the error that the operation would incur
                dt.error_recommended_operation_=event.data[1] 
                
                # update the global variables which keep track of the selected index to delete and the similar pair to merge
                # and the errors that would be incurred if they are committed
                dt.error_index_delete_={event.data[2]: event.data[1]}
                
                if dt.similar_pairs_:
                    
                    dt.error_similar_pair_={dt.similar_pairs_[0]: (event.data[6], event.data[7])}
                
            
            elif event.data[0]=="a":
                self.LPanel.operation.SetLabel("Merge with AND")
                
                # add one to each entry of the tuple to reflect the labels of the grid since
                # the grid numbering starts with 1
                first_factor=event.data[2][0]+1
                second_factor=event.data[2][1]+1
                
                self.LPanel.operation_factors.SetLabel(str((first_factor, second_factor)))
                self.LPanel.operation_error.SetLabel(str(event.data[1]))
                
                # update the user choice and merge operation
                dt.user_choose_=event.data[0]
                
                # update the operation required
                dt.merging_operation_="a"
                
                # update the similar pair to merge global variable
                dt.factor_pair_merge_=event.data[2]
                
                # update the error that the operation would incur
                dt.error_recommended_operation_=event.data[1]
                
                # update the global variables which keep track of the selected index to delete 
                # and the similar pair to merge and the errors that would be incurred if they are committed
                dt.error_index_delete_={dt.factoritems_.index(dt.ordered_factors_items_[0][0]): event.data[7]}
                dt.error_similar_pair_={dt.similar_pairs_[0]: (event.data[6], event.data[1])}
                
                # update the temp global variables in order to proceed with the plotting
                dt.clitems_temp= event.data[3]
                dt.trans_sets_temp= event.data[4]
                dt.factoritems_temp= event.data[5]
                
                # update the scores and number of 1s to reflect the merge
                # for that call the updates_OR_AND in the Constraints_new module
                updates_AND=cnstr.updates_OR_AND(dt.clitems_temp, dt.trans_sets_temp, \
                                                 dt.factoritems_, dt.factoritems_temp, dt.scores_factors_,\
                                                 dt.number_1s_,dt.reduced_index_cover_, "a", event.data[2])
                
                dt.scores_factors_temp= updates_AND[0]
                dt.number_1s_temp=updates_AND[1]
                
            else:
                self.LPanel.operation.SetLabel("Merge with OR") 
                
                # add one to each entry of the tuple to reflect the labels of the grid since
                # the grid numbering starts with 1
                first_factor=event.data[2][0]+1
                second_factor=event.data[2][1]+1
                
                self.LPanel.operation_factors.SetLabel(str((first_factor, second_factor)))
                self.LPanel.operation_error.SetLabel(str(event.data[1]))
                
                # update the user choice and merge operation
                dt.user_choose_=event.data[0]
                
                # update the operation required
                dt.merging_operation_="o"
                
                # update the similar pair to merge global variable
                dt.factor_pair_merge_=event.data[2]
                
                # update the error that the operation would incur
                dt.error_recommended_operation_=event.data[1]
                
                # update the global variables which keep track of the selected index to delete 
                # and the similar pair to merge and the errors that would be incurred if they are committed
                dt.error_index_delete_={dt.factoritems_.index(dt.ordered_factors_items_[0][0]): event.data[7]}
                dt.error_similar_pair_={dt.similar_pairs_[0]: (event.data[1], event.data[6])}
                
                # update the temp global variables in order to proceed with the plotting
                dt.clitems_temp= event.data[3]
                dt.trans_sets_temp= event.data[4]
                dt.factoritems_temp= event.data[5]
                
                # update the scores and number of 1s to reflect the merge
                # for that call the updates_OR_AND in the Constraints_new module
                updates_AND=cnstr.updates_OR_AND(dt.clitems_temp, dt.trans_sets_temp, \
                                                 dt.factoritems_, dt.factoritems_temp, dt.scores_factors_,\
                                                 dt.number_1s_,dt.reduced_index_cover_, "o", event.data[2])
                
                dt.scores_factors_temp= updates_AND[0]
                dt.number_1s_temp=updates_AND[1]
                
        # activate the Skip and Commit buttons since there is a recommended operation
        self.LPanel.skip_btn.Enable()
        self.LPanel.commit_btn.Enable()
        self.LPanel.visualize_btn.Enable()
         
        # the worker is done
        self.worker= None
        
        ####
        # THE OTHER EVENT HANDLERS
        ####
    
    # this function is almost the same as the OnSkip function in the plot_GUI module (commented out)
    def OnSkip(self, event):
        
        global skip_count
        
        # increase the skip count
        skip_count+=1
        
        # if the number of factors is 1 then stop since the system cannot recommend any further operation 
        if len(dt.factoritems_)>1: 
               
            # check if the user has not skipped the recommended delete operations for all factor items
               
            if skip_count!=len(dt.factoritems_)-1:
    
                # when the user clicks this button, we skip the current recommended operation and recomend the next best operation
                if dt.user_choose_=="d":
                    # recommend the next operation with the least error
                    # we get the error that would be incurred if the next factor item which covers the less number of 1s is deleted
                    next_index_to_delete=dt.factoritems_.index(dt.ordered_factors_items_[skip_count][0])
                    error_next_index_to_delete=dt.ordered_factors_items_[skip_count][1]
                    
                    # check if the similar_pairs global variable is not empty. If it is not empty proceed 
                    if dt.similar_pairs_:
                        
                        # compare the error of the delete operation to the other errors of the merge with OR and AND for the first similar pair
                        # get the errors for Merge with OR and AND
                        for k in dt.error_similar_pair_.keys():
                            similar_pair=k
                            error_merge_pair_or=dt.error_similar_pair_[k][0]
                            error_merge_pair_and=dt.error_similar_pair_[k][1]
                            
                            if error_next_index_to_delete<=error_merge_pair_or:
                                      
                                    if error_next_index_to_delete<=error_merge_pair_and:
                                        # the next operation in this case is a Delete operation
                                        # update the error that the recommended operation would incur
                                        dt.error_recommended_operation_=error_next_index_to_delete
                                        # update the factor index recommended to delete
                                        dt.factorindx_=next_index_to_delete
                                         
                                        # update the operation recommended to the user
                                        dt.user_choose_="d"
                                         
                                        # save error and index to the global variable
                                        dt.error_index_delete={next_index_to_delete:error_next_index_to_delete}
                     
                                        # set the corresponding labels
                                        self.LPanel.operation.SetLabel("Delete")
                                        self.LPanel.operation_factors.SetLabel(str(dt.factorindx_+1))
                                        self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))

                                        
                                    else:
                                        # the next operation is a merge with AND operation
                                        # update the error that the recommended operation would incur
                                        dt.error_recommended_operation_=error_merge_pair_and
                                        # update the factor pair recommended to merge
                                        dt.factor_pair_merge_=similar_pair
                                        # update the operation recommended to the user
                                        dt.user_choose_="a"
                                        
                                        # update the operation required
                                        dt.merging_operation_="a"
                                        
                                        # save error and index to the global variable
                                        dt.error_similar_pair_={similar_pair: (error_merge_pair_or, error_merge_pair_and)}
                                        
                                        # update the temp global variables to be able to do the plotting and the commit afterwards
                                        # for that we call the error_OR_AND from the Constraints_new module which does that
                                        temp_variables=cnstr.error_OR_AND(similar_pair, dt.clitems_, dt.trans_sets_, dt.factoritems_)
                                        
                                        dt.clitems_temp = temp_variables[5]
                                        dt.trans_sets_temp = temp_variables[6]
                                        dt.factoritems_temp = temp_variables[7]
                                        
                                        # update the scores and number of 1s to reflect the merge
                                        # for that call the updates_OR_AND in the Constraints_new module
                                        updates_AND=cnstr.updates_OR_AND(dt.clitems_temp, dt.trans_sets_temp, \
                                                                         dt.factoritems_, dt.factoritems_temp, dt.scores_factors_,\
                                                                         dt.number_1s_,dt.reduced_index_cover_, "a", similar_pair)
                                        
                                        dt.scores_factors_temp= updates_AND[0]
                                        dt.number_1s_temp=updates_AND[1]
                
                                        # set the corresponding labels
                                        self.LPanel.operation.SetLabel("Merge with AND")
                                        # add one to each entry of the tuple to reflect the labels of the grid since
                                        # the grid numbering starts with 1
                                        first_factor=similar_pair[0]+1
                                        second_factor=similar_pair[1]+1
                                        
                                        self.LPanel.operation_factors.SetLabel(str((first_factor, second_factor)))
                                        self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                                        
                            else:
                                if error_merge_pair_or<=error_merge_pair_and:
                                    
                                    # update the error that the recommended operation would incur
                                    dt.error_recommended_operation_=error_merge_pair_or
                                    # update the factor pair recommended to merge
                                    dt.factor_pair_merge_=similar_pair
                                    # update the operation recommended to the user
                                    dt.user_choose_="o"
                                    
                                    # update the operation required
                                    dt.merging_operation_="o"
                
                                    # save error and index to the global variable
                                    dt.error_similar_pair_={similar_pair: (error_merge_pair_or, error_merge_pair_and)}
                                    
                                    # update the temp global variables to be able to do the plotting and the commit afterwards
                                    # for that we call the error_OR_AND from the Constraints_new module which does that
                                    temp_variables=cnstr.error_OR_AND(similar_pair, dt.clitems_, dt.trans_sets_, dt.factoritems_)
                                    
                                    dt.clitems_temp = temp_variables[2]
                                    dt.trans_sets_temp = temp_variables[3]
                                    dt.factoritems_temp = temp_variables[4]
                                    
                                    # update the scores and number of 1s to reflect the merge
                                    # for that call the updates_OR_AND in the Constraints_new module
                                    updates_AND=cnstr.updates_OR_AND(dt.clitems_temp, dt.trans_sets_temp, \
                                                                     dt.factoritems_, dt.factoritems_temp, dt.scores_factors_,\
                                                                     dt.number_1s_,dt.reduced_index_cover_, "o", similar_pair)
                                    
                                    dt.scores_factors_temp= updates_AND[0]
                                    dt.number_1s_temp=updates_AND[1]
                
                                    # set the corresponding labels
                                    self.LPanel.operation.SetLabel("Merge with OR")
                                    # add one to each entry of the tuple to reflect the labels of the grid since
                                    # the grid numbering starts with 1
                                    first_factor=similar_pair[0]+1
                                    second_factor=similar_pair[1]+1
                                    
                                    self.LPanel.operation_factors.SetLabel(str((first_factor, second_factor)))
                                    self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                                    
                                    
                                    
                                else:
                                    # the next operation is a merge with AND operation
                                    # update the error that the recommended operation would incur
                                    dt.error_recommended_operation_=error_merge_pair_and
                                    # update the factor pair recommended to merge
                                    dt.factor_pair_merge_=similar_pair
                                    # update the operation recommended to the user
                                    dt.user_choose_="a"
                                    
                                    # update the operation required
                                    dt.merging_operation_="a"
                
                                    # save error and index to the global variable
                                    dt.error_similar_pair_={similar_pair: (error_merge_pair_or, error_merge_pair_and)}
                                    
                                    # update the temp global variables to be able to do the plotting and the commit afterwards
                                    # for that we call the error_OR_AND from the Constraints_new module which does that
                                    temp_variables=cnstr.error_OR_AND(similar_pair, dt.clitems_, dt.trans_sets_, dt.factoritems_)
                                    
                                    dt.clitems_temp = temp_variables[5]
                                    dt.trans_sets_temp = temp_variables[6]
                                    dt.factoritems_temp = temp_variables[7]
                                    
                                    # update the scores and number of 1s to reflect the merge
                                    # for that call the updates_OR_AND in the Constraints_new module
                                    updates_AND=cnstr.updates_OR_AND(dt.clitems_temp, dt.trans_sets_temp, \
                                                                     dt.factoritems_, dt.factoritems_temp, dt.scores_factors_,\
                                                                     dt.number_1s_,dt.reduced_index_cover_, "a", similar_pair)
                                    
                                    dt.scores_factors_temp= updates_AND[0]
                                    dt.number_1s_temp=updates_AND[1]
                                    
                                    # set the corresponding labels
                                    self.LPanel.operation.SetLabel("Merge with AND")
                                    # add one to each entry of the tuple to reflect the labels of the grid since
                                    # the grid numbering starts with 1
                                    first_factor=similar_pair[0]+1
                                    second_factor=similar_pair[1]+1
                                    
                                    self.LPanel.operation_factors.SetLabel(str((first_factor, second_factor)))
                                    self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                                    
                    else:
                        # the next operation in this case is a Delete operation
                        # update the error that the recommended operation would incur
                        dt.error_recommended_operation_=error_next_index_to_delete
                        # update the factor index recommended to delete
                        dt.factorindx_=next_index_to_delete
                        
                        # update the operation recommended to the user
                        dt.user_choose_="d"
                        
                        # save error and index to the global variable
                        dt.error_index_delete={next_index_to_delete:error_next_index_to_delete}
                                        
                        # set the corresponding labels
                        self.LPanel.operation.SetLabel("Delete")
                        self.LPanel.operation_factors.SetLabel(str(dt.factorindx_+1))
                        self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                       
                              
                                    
                # check the kind of operation previously recommended to the user
                # the user was recommended a Merge with AND operation but decided to skip    
                elif dt.user_choose_=="a":
                    # Check whether we could propose the merge with OR of the same pair or the delete operation
                    # check the factors in terms of increasing scores 
                    if dt.ordered_factors_items_[skip_count-1] not in dt.error_index_delete_.keys():
                        index_to_delete=dt.factoritems_.index(dt.ordered_factors_items_[skip_count-1][0])
                        error_index_to_delete=dt.ordered_factors_items_[skip_count-1][1]
                        
                    # now check the error of merging the same factor pair with OR
                    similar_pair=dt.factor_pair_merge_
                    
                    # error of merging pair using OR
                    error_merge_pair_or=dt.error_similar_pair_[similar_pair][0]
                    
                    if error_index_to_delete<=error_merge_pair_or:
                        
                        # the next operation in this case is a Delete operation
                        # update the error that the recommended operation would incur
                        dt.error_recommended_operation_=error_index_to_delete
                        # update the factor index recommended to delete
                        dt.factorindx_=index_to_delete
                        
                        # update the operation recommended to the user
                        dt.user_choose_="d"
                        
                        # save error and index to the global variable
                        dt.error_index_delete={index_to_delete:error_index_to_delete}
                                        
                        # set the corresponding labels
                        self.LPanel.operation.SetLabel("Delete")
                        self.LPanel.operation_factors.SetLabel(str(dt.factorindx_+1))
                        self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                        
                    else:
                        # update the error that the recommended operation would incur
                        dt.error_recommended_operation_=error_merge_pair_or
                        
                        # update the operation recommended to the user
                        dt.user_choose_="o"
                        
                        # update the operation required
                        dt.merging_operation_="o"
                
                        # save error and index to the global variable
                        dt.error_similar_pair_={similar_pair: (error_merge_pair_or, error_merge_pair_and)}
                        
                        # update the temp global variables to be able to do the plotting and the commit afterwards
                        # for that we call the error_OR_AND from the Constraints_new module which does that
                        temp_variables=cnstr.error_OR_AND(similar_pair, dt.clitems_, dt.trans_sets_, dt.factoritems_)
                        
                        dt.clitems_temp = temp_variables[2]
                        dt.trans_sets_temp = temp_variables[3]
                        dt.factoritems_temp = temp_variables[4]
                        
                        # update the scores and number of 1s to reflect the merge
                        # for that call the updates_OR_AND in the Constraints_new module
                        updates_AND=cnstr.updates_OR_AND(dt.clitems_temp, dt.trans_sets_temp, \
                                                         dt.factoritems_, dt.factoritems_temp, dt.scores_factors_,\
                                                         dt.number_1s_,dt.reduced_index_cover_, "o", similar_pair)
                        
                        dt.scores_factors_temp= updates_AND[0]
                        dt.number_1s_temp=updates_AND[1]
                            
                        # set the corresponding labels
                        self.LPanel.operation.SetLabel("Merge with OR")
                        # add one to each entry of the tuple to reflect the labels of the grid since
                        # the grid numbering starts with 1
                        first_factor=similar_pair[0]+1
                        second_factor=similar_pair[1]+1
                        
                        self.LPanel.operation_factors.SetLabel(str((first_factor, second_factor)))
                        self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                            
                # check the kind of operation previously recommended to the user
                # the user was recommended a Merge with OR operation but decided to skip             
                else:
                    # Check whether we could propose the merge with AND of the same pair or the delete operation
                    # check the factors in terms of increasing scores 
                    if dt.ordered_factors_items_[skip_count-1] not in dt.error_index_delete_.keys():
                        index_to_delete=dt.factoritems_.index(dt.ordered_factors_items_[skip_count-1][0])
                        error_index_to_delete=dt.ordered_factors_items_[skip_count-1][1]
                        
                    # now check the error of merging the same factor pair with OR
                    similar_pair=dt.factor_pair_merge_
                    
                    # error of merging pair using OR
                    error_merge_pair_or=dt.error_similar_pair_[similar_pair][0]
                    error_merge_pair_and=dt.error_similar_pair_[similar_pair][1]
                    
                    if error_index_to_delete<=error_merge_pair_and:
                        # the next operation in this case is a Delete operation
                        # update the error that the recommended operation would incur
                        dt.error_recommended_operation_=error_index_to_delete
                        # update the factor index recommended to delete
                        dt.factorindx_=index_to_delete
                        
                        # update the operation recommended to the user
                        dt.user_choose_="d"
                        
                        # save error and index to the global variable
                        dt.error_index_delete={index_to_delete:error_index_to_delete}
                                        
                        # set the corresponding labels
                        self.LPanel.operation.SetLabel("Delete")
                        self.LPanel.operation_factors.SetLabel(str(dt.factorindx_+1))
                        self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                    
                    else:
                        # update the error that the recommended operation would incur
                        dt.error_recommended_operation_=error_merge_pair_or
                       
                        # update the operation recommended to the user
                        dt.user_choose_="a"
                        
                        # update the operation required
                        dt.merging_operation_="a"
                
                        # save error and index to the global variable
                        dt.error_similar_pair_={similar_pair: (error_merge_pair_or, error_merge_pair_and)}
                        
                        # update the temp global variables to be able to do the plotting and the commit afterwards
                        # for that we call the error_OR_AND from the Constraints_new module which does that
                        temp_variables=cnstr.error_OR_AND(similar_pair, dt.clitems_, dt.trans_sets_, dt.factoritems_)
                        
                        dt.clitems_temp = temp_variables[5]
                        dt.trans_sets_temp = temp_variables[6]
                        dt.factoritems_temp = temp_variables[7]
                        
                        # update the scores and number of 1s to reflect the merge
                        # for that call the updates_OR_AND in the Constraints_new module
                        updates_AND=cnstr.updates_OR_AND(dt.clitems_temp, dt.trans_sets_temp, \
                                                         dt.factoritems_, dt.factoritems_temp, dt.scores_factors_,\
                                                         dt.number_1s_,dt.reduced_index_cover_, "a", similar_pair)
                        
                        dt.scores_factors_temp= updates_AND[0]
                        dt.number_1s_temp=updates_AND[1]
                
                        # set the corresponding labels
                        self.LPanel.operation.SetLabel("Merge with AND")
                        # add one to each entry of the tuple to reflect the labels of the grid since
                        # the grid numbering starts with 1
                        first_factor=similar_pair[0]+1
                        second_factor=similar_pair[1]+1
                        
                        self.LPanel.operation_factors.SetLabel(str((first_factor, second_factor)))
                        self.LPanel.operation_error.SetLabel(str(dt.error_recommended_operation_))
                    
            else:
                end_msg=wx.MessageDialog(None, "You have skipped all recommended operations. Click recommend to start all over!", "Information", wx.OK|wx.ICON_INFORMATION)
                
                if end_msg.ShowModal()==wx.ID_OK:
                        end_msg.Destroy()
                else:
                    end_msg.Destroy()
            
        else:
            end_msg=wx.MessageDialog(None, "It is left with only one factor.", "Information", wx.OK|wx.ICON_INFORMATION)
            
            if end_msg.ShowModal()==wx.ID_OK:
                        end_msg.Destroy()
            else:
                end_msg.Destroy()
                
                
    def OnMergeVisualize(self, event):
        
        global merge_pair
    
        if self.LPanel.Merge_AND_error.LabelText=="":
            
            msg_error=wx.MessageDialog(self, "Please select a pair to merge", "Error", wx.OK|wx.ICON_ERROR)
            
            if msg_error.ShowModal()==wx.ID_OK:
                msg_error.Destroy()
        else:
            # generate the data for the merging of the two using the OR operator
            trans,closed,factors=cnstr.merge_factor(merge_pair,dt.clitems_,dt.trans_sets_,dt.factoritems_)
            
            scores, number_1s=cnstr.updates_OR_AND(closed, trans, dt.factoritems_, factors, dt.scores_factors_, dt.number_1s_, dt.reduced_index_cover_, "o", merge_pair)
            
            # destroy previous window
            self.inst_plot_visual.Destroy()
            # plot global view
            self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, closed, trans, factors\
                                            , scores, factors.index(len(closed)-1), "d", "")
                
                    
    def OnVisualize(self, event):
        # when the user clicks this button he visualizes the recommended operation
        if dt.user_choose_=="d":
            # call the function to do the plotting (local view)
            self.PlotLocal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                            , dt.scores_factors_, dt.factorindx_, dt.user_choose_, "")
        else:
            self.PlotLocal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_temp, dt.trans_sets_temp, dt.factoritems_temp\
                                            , dt.scores_factors_temp, dt.factoritems_temp.index(len(dt.clitems_temp)-1), dt.user_choose_, dt.merging_operation_)
            
        
    def OnCommit (self, event):
        # commit the recommended operation
            
        # we update the global variables to reflect the changes
        if dt.user_choose_=="d":
            
            # remove factor from grid table
            self.grid.GetTable().RemoveData([dt.factorindx_])
            
            #update the factors
            new_factors=[x for x in dt.factoritems_ if x!=dt.factoritems_[dt.factorindx_]]
        
            # factor in delete list
            delete_list_factor=[dt.factoritems_[dt.factorindx_]]
            # update the scores
            # by deleting some factors, the number of ones covered alone by some other factors may increase
            # so we look through the reduced_cover and update the scores accordingly
            for k in dt.reduced_index_cover_.keys():
                # we only consider the indices which are covered only by 2 factors since these will be covered by only 1 if the other is deleted
                if len(dt.reduced_index_cover_[k])==2:
                      
                    factor_to_update=list(set(dt.reduced_index_cover_[k]).difference(set(delete_list_factor)))
                      
                    # if the difference is not empty then it contains the factor whose score should be incremented by 1
                    if len(factor_to_update)==1:
                        if factor_to_update[0] in dt.scores_factors_.keys():
                            
                            dt.scores_factors_[factor_to_update[0]]+=1
                            # update the unique area and error of this factor from the data in the grid
                            self.grid.GetTable().SetValue(new_factors.index(factor_to_update[0]), 5, dt.scores_factors_[factor_to_update[0]])
                            self.grid.GetTable().SetValue(new_factors.index(factor_to_update[0]), 6, dt.scores_factors_[factor_to_update[0]]) 
            
            # update the accumulated error
            dt.acc_error_+=dt.error_recommended_operation_
            
            # remove the factors deleted from the scores
            dt.scores_factors_={i: dt.scores_factors_[i] for i in dt.scores_factors_.keys() if i not in delete_list_factor}
            
            # update factors
            dt.factoritems_= new_factors
            
            # update the statistics portion of the interface
            self.LPanel.current_nbfact.SetLabel(str(len(dt.factoritems_)))
            self.LPanel.acc_error.SetLabel(str(dt.acc_error_))
            
            # empty the recommendations labels
            self.LPanel.operation_error.SetLabel("")
            self.LPanel.operation_factors.SetLabel("")
            self.LPanel.operation.SetLabel("")
            
            # update global view
            # destroy previous window
            self.inst_plot_visual.Destroy()
            # plot global view
            self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                            , dt.scores_factors_, 0, "d", "")
            
            
        else:
            
            # first we delete the factors in the merge pair from the grid
            self.grid.GetTable().RemoveData(list(dt.factor_pair_merge_))
            
            # update all the global variables using the temp global variables
            dt.clitems_= dt.clitems_temp
            dt.trans_sets_= dt.trans_sets_temp
            dt.factoritems_ = dt.factoritems_temp
            dt.scores_factors_= dt.scores_factors_temp
            dt.number_1s_ = dt.number_1s_temp
            
            if dt.scores_factors_[dt.factoritems_[-1]]!=0:
                # generate new data
                # get the row and column names of the factor resulting from the merge
                rows_names_merged, nb_trans, col_names_merged, nb_items=row_col_factor(dt.clitems_, dt.trans_sets_, dt.factoritems_[-1]) 
                
                # new row to add to the grid
                data=()
                new_data_row=(col_names_merged, nb_items, rows_names_merged, nb_trans, dt.number_1s_[dt.factoritems_[-1]], dt.scores_factors_[dt.factoritems_[-1]], dt.scores_factors_[dt.factoritems_[-1]],)
                data = data + (new_data_row,)
                
                # call the AddData method of the pygridTablebase class to update the grid
                self.grid.GetTable().AddData(data)
                
            else:
                # since the unique area of the new factor is zero, we remove it from the list of factors
                dt.factoritems_=dt.factoritems_[:-1]
                
            # update accumulated error
            dt.acc_error_+=dt.error_recommended_operation_
            
            # update the statistics portion of the interface
            self.LPanel.current_nbfact.SetLabel(str(len(dt.factoritems_)))
            self.LPanel.acc_error.SetLabel(str(dt.acc_error_))
                
            # update global view
            # destroy previous window
            self.inst_plot_visual.Destroy()
            # plot global view
            self.PlotGlobal(dt.indices_1s_, dt.number_columns_, dt.number_rows_, dt.clitems_, dt.trans_sets_, dt.factoritems_\
                                            , dt.scores_factors_, 0, "d", "")
                                    
    def closewindow(self,event):
        self.Destroy()
    
    ####
    # PLOTTING FUNCTIONS
    ####
    # plot global view (view of the data with the selected factor highlighted)
    def PlotGlobal(self, indices_1s, nb_cols, nb_rows, closed, trans, factors, scores, indx, user_choice, merging_operation):
        
        # close previous window
        
        
        #generate full matrix by adding (element addition) the factors to each other
        dt.full_mat_=vs.gen_matrix(closed, trans, factors, nb_cols, nb_rows)
        
        self.inst_plot_visual=visualization.plotFrame(dt.full_mat_, closed, trans,\
                                                 factors, scores, indices_1s,\
                                                 indx,"",user_choice,merging_operation)
        # update the global variables in the important_data module to be able to access them somewhere else so we need not
        dt.factorindx_=indx
        dt.add_info_=""
        dt.user_choose_=user_choice
        dt.merging_operation_=merging_operation
        
        # call function to generate the matrix to plot using imshow    
        ind,temp_perm_mat,perm_row,perm_mat=self.inst_plot_visual.generate_matrices_to_plot()
        
        # update more global variables
        dt.indices_=ind
        dt.temp_perm_mat_=temp_perm_mat
        dt.perm_row_=perm_row
        dt.perm_mat_=perm_mat
        
        # plot the matrix-- global view
        self.inst_plot_visual.draw_figure(ind,perm_row, perm_mat)
        
        # show the plot 
        self.inst_plot_visual.Show()
    
    # plot local view (view of the factor)
    def PlotLocal(self, indices_1s, nb_cols, nb_rows, closed, trans, factors, scores, indx, user_choice, merging_operation):
        #generate full matrix by adding (element addition) the factors to each other
        dt.full_mat_=vs.gen_matrix(closed, trans, factors, nb_cols, nb_rows)
        
        inst_plot_visual2=visualization.plotFrame(dt.full_mat_, closed, trans,\
                                                 factors, scores, indices_1s,\
                                                 indx,"",user_choice,merging_operation)
        
        # update the global variables in the important_data module to be able to access them somewhere else so we need not
        dt.factorindx_=indx
        dt.add_info_=""
        dt.user_choose_=user_choice
        dt.merging_operation_=merging_operation
                
        # call function to generate the matrix to plot using imshow    
        ind,temp_perm_mat,perm_row,perm_mat=inst_plot_visual2.generate_matrices_to_plot()
        
        # update more global variables
        dt.indices_=ind
        dt.temp_perm_mat_=temp_perm_mat
        dt.perm_row_=perm_row
        dt.perm_mat_=perm_mat
        
        # plot the matrix-- global view
        inst_plot_visual2.draw_figure_local(ind,temp_perm_mat,perm_row, perm_mat,0)
        
        # show the plot 
        inst_plot_visual2.Show(True)
    

####    
# RUN THE FRAME
####
if __name__=='__main__':
    app=wx.App(False)
    frame=interactive_BMF()
    frame.Show()
    app.MainLoop()
                
        
        
    