'''
Created on Mar 10, 2016

@author: mukuze
'''
import wx
# from wx.lib.pubsub import pub
import numpy as np
from copy import deepcopy
import pandas as pd
import os
from textwrap import wrap
 
import matplotlib
matplotlib.use('WXAgg')
from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigCanvas, NavigationToolbar2WxAgg as NavigationToolbar
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec



# import other modules
import Visualization_functions as vs

# import the module containing the global variables that we want to share across modules
import important_data as dt


# global variable
count=0
# global variable to count how many times the user has rejected the recommended operation
skip_count=0

class plotFrame(wx.Frame):
    
    title="View of the current factor"
    
    # the data
    full_mat=[]
    clitems=[]
    trans_sets=[]
    factoritems=[]
    scores_factors={}
    indices_1s=[]
    factorindx=0
    add_info=""
    user_choose=""
    merging_operation=""
    
    def __init__(self,input_mat,closed,trans,factors,scores,indices_,fact_indx,info,choice,merg_op):
        wx.Frame.__init__(self,None,-1,self.title)
        
        
        # initialize the data
        self.full_mat=input_mat
        self.clitems=closed
        self.trans_sets=trans
        self.factoritems=factors
        self.scores_factors=scores
        self.indices_1s=indices_
        self.factorindx=fact_indx
        self.add_info=info
        self.user_choose=choice
        self.merging_operation=merg_op
 
        self.create_menu()
        self.create_status_bar()
        
        # as we initialize the panel we label the Delete recommended factor or accept merge button accordingly
        if self.merging_operation=="":
            label="Delete recommended factor"
        else:
            label="Accept recommended factor merge"
            
        self.create_main_panel(label)
        
        # when there is only one factor item left, we disable the delete/Merge button since the system has come to an end
        if len(dt.factoritems_)==1:
            self.del_mergebutton.Disable()
            self.skipbutton.Disable()

    def create_menu(self):
        
        self.menubar=wx.MenuBar()
        
        menu_file=wx.Menu()
        
        m_save=menu_file.Append(-1,"&Save plot\tCtrl-S","Save plot to file")
        self.Bind(wx.EVT_MENU, self.OnSavePlot, m_save)
        
        menu_file.AppendSeparator()
        
        m_exit=menu_file.Append(-1,"E&xit\tCtrl-X","Exit")
        self.Bind(wx.EVT_MENU, self.OnExit, m_exit)
        
        menu_help=wx.Menu()
        
        m_about=menu_help.Append(-1,"&About\tF1","about the app")
        self.Bind(wx.EVT_MENU, self.OnAbout, m_about)
        
        self.menubar.Append(menu_file, "&File")
        self.menubar.Append(menu_help, "&Help")
        self.SetMenuBar(self.menubar)
        
    def create_main_panel(self, label):
        
        # create panel
        self.panel=wx.Panel(self)
        
        # create figure and canvas
        self.dpi=100
        self.fig= Figure ((6.0, 5.0), dpi=self.dpi)
        self.canvas= FigCanvas(self.panel, -1, self.fig)

        self.canvas.Update()
        self.canvas.Refresh()
        self.Update()
        
        
        self.nextbutton=wx.Button(self.panel, -1, "Next")
        self.Bind(wx.EVT_BUTTON, self.OnNextButton, self.nextbutton)
        
        self.previousbutton=wx.Button(self.panel, -1, "Previous")
        self.Bind(wx.EVT_BUTTON, self.OnPreviousButton, self.previousbutton)
        
        
        self.nextbutton.SetToolTipString('See the next group of 25 rows covered by the current factor')
        self.previousbutton.SetToolTipString('See the previous group of 25 rows covered by the current factor')

        self.toolbar=NavigationToolbar(self.canvas)
        
        # Layout with box sizers
        
        self.vbox=wx.BoxSizer(wx.VERTICAL)
        self.vbox.Add(self.canvas, 1, wx.LEFT | wx.TOP | wx.GROW)
        self.vbox.Add(self.toolbar, 0, wx.EXPAND)
        self.vbox.AddSpacer(10)
        
        self.hbox=wx.BoxSizer(wx.HORIZONTAL)
        flags=wx.ALIGN_LEFT | wx.ALL | wx.ALIGN_CENTER_VERTICAL
        self.hbox.Add(self.previousbutton, 0, border=3, flag=flags)
        self.hbox.Add(self.nextbutton, 0, border=3, flag=flags)

        self.hbox.AddSpacer(30)
        
        
        self.vbox.Add(self.hbox, 0, flag=wx.ALIGN_LEFT | wx.TOP)
        
        self.panel.SetSizer(self.vbox)
        self.vbox.Fit(self)
        
    def create_status_bar(self):
        self.statusbar=self.CreateStatusBar()
    
    def generate_matrices_to_plot(self):
        # generate the matrix for plot
        # get the indices of 1s in full_data
        orig_indx=self.indices_1s
        current_indx=np.transpose(np.nonzero(self.full_mat))
         
        # we do the checks and makes the appropriate changes in case we are merging
        # using either OR or AND
        # indices covered by the current item set in the current full data
        indices_itemset=[(x,y) for x in self.trans_sets[self.factoritems[self.factorindx]] for y in self.clitems[self.factoritems[self.factorindx]]]
        # indices covered by the current itemset in the original full data
        indx_new=[x for x in orig_indx if x[0] in self.trans_sets[self.factoritems[self.factorindx]]]
        # check whether we are doing merge with OR [o] or AND [a] operator
        if self.merging_operation=="o":
            # indices that are turned into 1 due to the OR operation
            indices_set=set(map(tuple,indices_itemset)).difference(map(tuple,indx_new))
            indices=map(list,indices_set)
     
            # assign the 0's that turned into 1's a value (0.5)
            for x in range(len(indices)):
                self.full_mat[indices[x][0],indices[x][1]]=0.5
    #         print full_data
             
        if self.merging_operation=="a":
            indx_new_current=[x for x in current_indx if x[0] in self.trans_sets[self.factoritems[self.factorindx]]]
            indices_set_=set(map(tuple,indx_new)).difference(map(tuple,indx_new_current))
            indices_=map(list,indices_set_)
            # assign the 1's that turned into 0's a value (0.25)
            for x in range(len(indices_)):
                self.full_mat[indices_[x][0],indices_[x][1]]=0.25
         
        # we check the indices originally on the rows where the factor item is present
        # and return the indices which are covered but not by this factor
        # they were uncovered by the AND merge operation
        indices_set__=set(map(tuple,indx_new)).difference(map(tuple,indices_itemset))
        indices__=map(list,indices_set__)
         
        # assign the 1's which are on the same rows but covered by other factors 
        # meaning not covered by this factor 0.7 which will be changed to -1 later on in the code 
        for x in range(len(indices__)):
            if self.full_mat[indices__[x][0],indices__[x][1]]==1:
                self.full_mat[indices__[x][0],indices__[x][1]]=0.7
         
        perm_mat, perm_row=vs.permute_rows(self.full_mat,self.trans_sets,self.factoritems,self.factorindx)
        temp_perm_mat=deepcopy(perm_mat)  
        indx_= vs.sort_indx(temp_perm_mat)
         
        # plot back into full mat so we get the factor_item in left top corner and sorted
        temp_perm_mat[:len(self.trans_sets[self.factoritems[self.factorindx]]),:].sort(axis=1)
        temp_perm_mat[:len(self.trans_sets[self.factoritems[self.factorindx]]),:]=temp_perm_mat[:len(self.trans_sets[self.factoritems[self.factorindx]]),:][:,::-1] 
         
        # assign the 1's which are on the same rows but covered by other factors 
        # meaning not covered by this factor -1
        for x in range(len(self.trans_sets[self.factoritems[self.factorindx]])):
            arr_=perm_mat[x]
            arr_[arr_==0.7]=-1
            perm_mat[x]=arr_
         
        # Modify the full matrix for plotting so that the 1's that are covered by other
        # factor_items are given -1 and we use gray in the color map to plot them
        for x in range(len(self.trans_sets[self.factoritems[self.factorindx]]),len(perm_mat)):
            arr=perm_mat[x]
            arr[arr>0]=-1
            perm_mat[x]=arr
        
        return indx_,temp_perm_mat,perm_row,perm_mat
    
    def draw_figure(self,indx_,perm_row,perm_mat):
        
        # disable the next, previous, delete buttons
        self.nextbutton.Disable()
        self.previousbutton.Disable()
#         self.del_mergebutton.Disable()
        
        # divide the subplot into 2 parts, one part to plot the data (global view) and the other to plot the colorbar
        self.gs = gridspec.GridSpec(1, 2,width_ratios=[30,1])
        self.gs.update(wspace=0.05)
        # create axes and subplot
        # For the global view
        self.ax1 = self.fig.add_subplot(self.gs[0])
        # For the colorbar
        self.ax2=self.fig.add_subplot(self.gs[1])
        

        if self.add_info=="":
            infos="factor_item at index:"+str(self.factorindx)+"\n"+"factor_item number:"+str(self.factoritems[self.factorindx])+"\n"+"number of 1s covered alone [red color]:"+str(str(self.scores_factors[self.factoritems[self.factorindx]]))
        else:
            infos=self.add_info
     
        if len(perm_mat)<10000:
     
            plt.rcParams["axes.titlesize"] = 8
            # define the colormaps for both the factor_item and the whole data
            cmap = matplotlib.colors.ListedColormap(['0.75','w','m','b','r', 'y','k'])

            bounds=[-1.,0.,0.25,0.5,1.,2.,3.,4.]
             
            norm= matplotlib.colors.BoundaryNorm(bounds, cmap.N)

            # clear the axes
            self.ax1.clear()
          
            img=self.ax1.imshow(perm_mat, interpolation='nearest', cmap=cmap, norm=norm)
            self.ax1.set_title(infos,fontsize='8')
            # do not show any labels for the axes
            self.ax1.set_xticks([])
            self.ax1.set_yticks([])
            self.ax1.set_aspect('auto')
            # make a colorbar
            cbar=plt.colorbar(img, cmap=cmap,cax=self.ax2)
            labels=['Covered by other factors','Uncovered','Covered that become uncovered (due to AND merge)','Uncovered that become covered (due to OR merge)','Covered by this factor alone','covered by this factor and any other one factor','covered by this factor and at least any other two']
            loc=[-0.5,0.125,0.375,0.75,1.5,2.5,3.5]
            cbar.set_ticks(loc)
            cbar.ax.set_yticklabels(['\n'.join(wrap(l, 11)) for l in labels])
            cbar.ax.tick_params(labelsize=6) 
            
            self.canvas.draw()
            
        else:
            user_dialog=wx.MessageDialog(None, "Data too big to plot", "Information",wx.OK|wx.ICON_INFORMATION)
            if user_dialog.ShowModal()==wx.ID_OK:
                self.closewindow(wx.EVT_CLOSE)
             
        
    def draw_figure_local(self,indx_,temp_perm_mat,perm_row,perm_mat,cnt):
        global count
            
        if self.add_info=="":
            infos="factor_item at index:"+str(self.factorindx)+"\n"+"factor_item number:"+str(self.factoritems[self.factorindx])+"\n"+"number of 1s covered alone [red color]:"+str(str(self.scores_factors[self.factoritems[self.factorindx]]))
        else:
            infos=self.add_info
        # Data with the 1's covered by the factor_item
        factor_mat=temp_perm_mat[0:len(self.trans_sets[self.factoritems[self.factorindx]]), 0:len(self.clitems[self.factoritems[self.factorindx]])]
        
        # put the data into a panda dataframe
        rows_cols_names=vs.get_col_row_names()
        
        count=cnt
        
        nb_iterations= int((len(factor_mat)/25)+1)
        
        # divide the subplot into 1 part, we are not plotting the colorbar
        self.gs1 = gridspec.GridSpec(25, 1,width_ratios=[30,1])
        
        for n_ in range(nb_iterations):
            
            plt.rcParams["axes.titlesize"] = 12
#             plt.rcParams.update({'figure.autolayout': True})
            # define the colormaps for both the factor_item and the whole data
            cmap = matplotlib.colors.ListedColormap(['0.75','w','m','b','r','y','k'])
            # create a normalize object the describes the limits of
            # each color
            bounds=[-1.,0.,0.25,0.5,1.,2.,3.,4.]
             
            norm= matplotlib.colors.BoundaryNorm(bounds, cmap.N)

            for i in range(25):
                
                names_cols_f=[rows_cols_names[1][x] for x in self.clitems[self.factoritems[self.factorindx]]]
#                 names_cols_f=[rows_cols_names[1][x] for x in indx_[count+i][0:len(self.clitems[self.factoritems[self.factorindx]])]]
#                 names_rows=[rows_cols_names[0][x] for x in perm_row]  
                factor_mat_headers=pd.DataFrame(columns=names_cols_f,index=[rows_cols_names[0][perm_row[count+i]]])
                factor_mat_headers.loc[rows_cols_names[0][perm_row[count+i]]] = factor_mat[count+i]
                
                # create axes
                # leave some space on the left to be able to see the long names and avoid truncating them
                if dt.show_ylabels==True:
                    self.fig.subplots_adjust(left=0.4)
                    
                self.ax3 = self.fig.add_subplot(self.gs1[i])
                
                
#                 img=self.ax3.imshow([factor_mat_1[count+i]], interpolation='nearest', cmap=cmap, norm=norm)
                img=self.ax3.imshow([factor_mat[count+i]], interpolation='nearest', cmap=cmap, norm=norm)
                
                # set the title with infos only once
                if i==0:
                    self.ax3.set_title(infos,fontsize='8')
                
                # set the sticks to empty
                self.ax3.set_xticks([])
                self.ax3.set_yticks([])
                if dt.show_ylabels==True:
                    self.ax3.set_yticks(np.arange(factor_mat_headers.shape[0]))
                    self.ax3.set_yticklabels(factor_mat_headers.index)
                    
                self.ax3.set_aspect('auto')
                
                if i==24:
                    # only set the sticks of the last row     
                    self.ax3.set_xticks(np.arange(factor_mat_headers.shape[1]))
                    self.ax3.set_xticklabels(names_cols_f)
                    
                if count+i==len(factor_mat)-1:
                    # only set the sticks of the last row     
                    self.ax3.set_xticks(np.arange(factor_mat_headers.shape[1]))
                    self.ax3.set_xticklabels(names_cols_f)
                    break
            
            # draw canvas
            self.canvas.draw()
            self.canvas.Update()
            self.canvas.Refresh()
            
            if nb_iterations>1:
                self.nextbutton.Enable()
                self.previousbutton.Enable()
            else:
                self.nextbutton.Disable()
                self.previousbutton.Disable()
            
            # At the end (when there are no more rows covered by the current factor to display) the next button should be disabled
            if count+i==len(factor_mat)-1:
                self.nextbutton.Disable()
            # At the start (when count is zero) the button should be disabled
            if count==0:
                self.previousbutton.Disable()
            
    def OnNextButton(self,event):
        # close the previous window
        self.Destroy()
        
        # enable the previous button
        self.previousbutton.Enable()
        
        # create an instance and call the next group of rows covered by the current factor
        global count
        instance=plotFrame(self.full_mat, self.clitems, self.trans_sets,\
                                                 self.factoritems, self.scores_factors, self.indices_1s,\
                                                 self.factorindx, self.add_info, self.user_choose, self.merging_operation)
        
        # increase the count, in order to show the next group of 25 rows affected by the factor if they are any    
        count=count+25
        
        # call the functions to make the drawing of the nent group of 25 rows affected by the factor if they are any
        instance.draw_figure_local(dt.indices_, dt.temp_perm_mat_, dt.perm_row_, dt.perm_mat_, count)
        instance.canvas.Update()
        instance.canvas.Refresh()
        instance.Show(True)

    def OnPreviousButton(self,event):
        # close the previous window
        self.Destroy()
        
        # create an instance and call the previous group of rows covered by the current factor
        global count
        instance_prev=plotFrame(self.full_mat, self.clitems, self.trans_sets,\
                                                 self.factoritems, self.scores_factors, self.indices_1s,\
                                                 self.factorindx, self.add_info, self.user_choose, self.merging_operation)
        
        # decrease the count, in order to show the next group of 25 rows affected by the factor if they are any    
        count=count-25
        instance_prev.draw_figure_local(dt.indices_, dt.temp_perm_mat_, dt.perm_row_, dt.perm_mat_, count)
        instance_prev.canvas.Update()
        instance_prev.canvas.Refresh()
        instance_prev.Show(True)
        
    def OnSavePlot(self,event):
        file_choices="PNG (*.png) |*.png"
        
        dlg=wx.FileDialog(self,message="Save plot as...", defaultDir=os.getcwd(), defaultFile="plot.png", wildcard=file_choices, style=wx.SAVE)
        
        if dlg.ShowModal()==wx.ID_OK:
            path=dlg.GetPath()
            self.canvas.print_figure(path, dpi=self.dpi)
            self.flash_status_message("Saved to %s" % path)
      

    def OnExit(self,event):
        self.Destroy()
        
    def OnAbout(self,event):
        msg="""An app for interactive Constrained Boolean matrix factorization:
        """
        
        dlg=wx.MessageDialog(self,msg,"About", wx.OK)
        dlg.ShowModal()
        dlg.Destroy()
        
    def flash_status_message(self,msg, flash_len_ms=1500):
            self.statusbar.SetStatusText(msg)
            self.timeroff=wx.Timer(self)
            self.Bind(wx.EVT_TIMER, self.OnFlashStatusOff, self.timeroff)
            self.timeroff.Start(flash_len_ms, oneShot=True)
            
    def OnFlashStatusOff(self,event):
        self.statusbar.SetStatusText('')
    
    def closewindow(self,event):
        self.Destroy()
        
if __name__=='__main__':
    app=wx.App(False)
    frame=plotFrame()
    frame.Show()
    app.MainLoop()
        